// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "InputManager.hpp"
#include "game/Game.hpp"

using namespace Dawn;

void InputManager::init(const std::shared_ptr<Game> game) {
  auto window = game->renderHost.window;
  
  glfwSetCursorPosCallback(window, [](
    GLFWwindow* window,
    double_t x,
    double_t y
  ) {
    auto game = (Game*)glfwGetWindowUserPointer(window);
    assertNotNull(game, "Game is not set on window user pointer.");
    game->inputManager.rawInputValues[INPUT_MANAGER_AXIS_MOUSE_X] = (float_t) x;
    game->inputManager.rawInputValues[INPUT_MANAGER_AXIS_MOUSE_Y] = (float_t) y;
  });

  glfwSetMouseButtonCallback(window, [](
    GLFWwindow* window,
    int32_t button,
    int32_t action,
    int32_t mods
  ) {
    auto game = (Game*)glfwGetWindowUserPointer(window);
    assertNotNull(game, "Game is not set on window user pointer.");
    switch(action) {
      case GLFW_PRESS:
        game->inputManager.rawInputValues[INPUT_MANAGER_AXIS_MOUSE_0 + button] = 1.0f;
        return;
      case GLFW_RELEASE:
        game->inputManager.rawInputValues[INPUT_MANAGER_AXIS_MOUSE_0 + button] = 0.0f;
        return;
      default:
        return;
    }
  });

  glfwSetKeyCallback(window, [](
    GLFWwindow* window,
    int32_t key,
    int32_t scancode,
    int32_t action,
    int32_t mods
  ) {
    auto game = (Game*)glfwGetWindowUserPointer(window);
    assertNotNull(game, "Game is not set on window user pointer.");

    switch(action) {
      case GLFW_PRESS:
        game->inputManager.rawInputValues[key] = 1.0f;
        return;
      case GLFW_RELEASE:
        game->inputManager.rawInputValues[key] = 0.0f;
        return;
      default:
        return;
    }
  });

  // Set default values
  this->bind(InputBind::Up, GLFW_KEY_W);
  this->bind(InputBind::Up, GLFW_KEY_UP);
  this->bind(InputBind::Down, GLFW_KEY_S);
  this->bind(InputBind::Down, GLFW_KEY_DOWN);
  this->bind(InputBind::Left, GLFW_KEY_A);
  this->bind(InputBind::Left, GLFW_KEY_LEFT);
  this->bind(InputBind::Right, GLFW_KEY_D);
  this->bind(InputBind::Right, GLFW_KEY_RIGHT);
  this->bind(InputBind::Run, GLFW_KEY_LEFT_SHIFT);
  this->bind(InputBind::Run, GLFW_KEY_RIGHT_SHIFT);
  this->bind(InputBind::Action, GLFW_KEY_SPACE);
  this->bind(InputBind::Action, GLFW_KEY_ENTER);
  this->bind(InputBind::Action, GLFW_KEY_E);
}

float_t InputManager::getInputValue(int32_t axis) {
  auto exist = rawInputValues.find(axis);
  if(exist == rawInputValues.end()) return 0.0f;
  return exist->second;
}

InputManager::~InputManager() {
  // Nothing to do here.
}