// Copyright (c) 2024 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "assert/assert.hpp"
#include "scene/SceneItem.hpp"
#include "Map.hpp"
#include "component/entity/Entity.hpp"

using namespace Dawn;

void Map::entityNotifyInit(std::shared_ptr<SceneItem> item) {
  auto entity = item->getComponent<Entity>();
  assertNotNull(entity, "Entity component not found on item.");
  auto id = entity->id;
  assertMapNotHasKey(this->entities, id, "Entity already exists in Map.");
  assertTrue(id != EntityID::Null, "Entity ID is invalid.");
  this->entities[id] = entity;
}

void Map::entityNotifyDispose(std::shared_ptr<SceneItem> item) {
  auto entity = item->getComponent<Entity>();
  assertNotNull(entity, "Entity component not found on item.");
  auto id = entity->id;
  assertMapHasKey(this->entities, id, "Entity does not exist in Map.");
  assertTrue(id != EntityID::Null, "Entity ID is invalid.");
  this->entities.erase(id);
}

void Map::onInit() {
}

void Map::onDispose() {
}

void Map::setSize(
  const uint32_t width,
  const uint32_t height,
  const uint32_t depth,
  const uint8_t chunkWidth,
  const uint8_t chunkHeight,
  const uint8_t chunkDepth
) {
  this->chunkOrder.clear();

  this->mapWidth = width;
  this->mapHeight = height;
  this->mapDepth = depth;
  this->mapChunkWidth = chunkWidth;
  this->mapChunkHeight = chunkHeight;
  this->mapChunkDepth = chunkDepth;
  this->mapChunkCount = chunkWidth * chunkHeight * chunkDepth;
  this->chunks = std::make_unique<std::shared_ptr<Chunk>[]>(
    this->mapChunkCount
  );

  this->chunkTopLeft.x = 0;
  this->chunkTopLeft.y = 0;
  this->chunkTopLeft.z = 0;
  
  // Load initial chunks
  for(auto z = 0; z < chunkDepth; z++) {
    for(auto y = 0; y < chunkHeight; y++) {
      for(auto x = 0; x < chunkWidth; x++) {
        auto chunk = std::make_shared<Chunk>();
        memset(chunk->tiles, 0, sizeof(chunk->tiles));
        chunk->position.x = x;
        chunk->position.y = y;
        chunk->position.z = z;
        this->chunks[x + y * chunkWidth + z * chunkWidth * chunkHeight] = chunk;
        
        this->chunkOrder.push_back(chunk);
      }
    }
  }
}

std::shared_ptr<World> Map::getWorld() {
  auto world = this->world.lock();
  assertNotNull(world, "World cannot be null?");
  return world;
}

std::shared_ptr<Entity> Map::getEntity(const EntityID id) {
  assertMapHasKey(this->entities, id, "Entity does not exist in Map.");
  auto ent = this->entities[id];
  auto lock = ent.lock();
  if(!lock) {
    this->entities.erase(id);
    return nullptr;
  }
  assertTrue(lock->id == id, "Entity ID mismatch.");
  return lock;
}

std::shared_ptr<Entity> Map::getEntityAt(const EntityTilePosition &pos) {
  for(auto &pair : this->entities) {
    auto ent = pair.second.lock();
    if(!ent) continue;
    if(ent->tilePosition != pos) continue;
    return ent;
  }
  return nullptr;
}

std::shared_ptr<Chunk> Map::getChunkAt(const struct ChunkPosition &pos) {
  assertTrue(pos.x >= this->chunkTopLeft.x, "Chunk X out of bounds.");
  assertTrue(pos.y >= this->chunkTopLeft.y, "Chunk Y out of bounds.");
  assertTrue(pos.z >= this->chunkTopLeft.z, "Chunk Z out of bounds.");

  // First, offset position from top left
  auto offset = pos - this->chunkTopLeft;

  assertTrue(offset.x < this->mapChunkWidth, "Chunk X out of bounds.");
  assertTrue(offset.y < this->mapChunkHeight, "Chunk Y out of bounds.");
  assertTrue(offset.z < this->mapChunkDepth, "Chunk Z out of bounds.");
  
  // Convert to index
  auto i = (
    offset.x +
    offset.y * this->mapChunkWidth +
    offset.z * this->mapChunkWidth * this->mapChunkHeight
  );

  assertTrue(i >= 0 && i < this->mapChunkCount, "Chunk index out of bounds.");
  return this->chunkOrder[i];
}

std::shared_ptr<Chunk> Map::getChunkAtTile(const EntityTilePosition &pos) {
  auto chunkPos = pos.toChunkPosition();
  return this->getChunkAt(chunkPos);
}

struct Tile Map::getTileAt(const EntityTilePosition &pos) {
  auto chunk = this->getChunkAtTile(pos);
  auto tilePos = pos.toTilePosition();
  return chunk->getTileAt(tilePos);
}